@@ -42,15 +42,6 @@ func NewMakeCommands() *cobra.Command {
 		Use:   "build",
 		Short: "Build delve",
 		Run: func(cmd *cobra.Command, args []string) {
-			tagFlag := prepareMacnative()
-			if len(*Tags) > 0 {
-				if len(tagFlag) == 0 {
-					tagFlag = "-tags="
-				} else {
-					tagFlag += ","
-				}
-				tagFlag += strings.Join(*Tags, ",")
-			}
 			envflags := []string{}
 			if len(Architecture) > 0 {
 				envflags = append(envflags, "GOARCH="+Architecture)
@@ -59,9 +50,9 @@ func NewMakeCommands() *cobra.Command {
 				envflags = append(envflags, "GOOS="+OS)
 			}
 			if len(envflags) > 0 {
-				executeEnv(envflags, "go", "build", "-ldflags", "-extldflags -static", tagFlag, buildFlags(), DelveMainPackagePath)
+				executeEnv(envflags, "go", "build", "-ldflags", "-extldflags -static", tagFlags(), buildFlags(), DelveMainPackagePath)
 			} else {
-				execute("go", "build", "-ldflags", "-extldflags -static", tagFlag, buildFlags(), DelveMainPackagePath)
+				execute("go", "build", "-ldflags", "-extldflags -static", tagFlags(), buildFlags(), DelveMainPackagePath)
 			}
 			if runtime.GOOS == "darwin" && os.Getenv("CERT") != "" && canMacnative() {
 				codesign("./dlv")
@@ -78,8 +69,7 @@ func NewMakeCommands() *cobra.Command {
 		Use:   "install",
 		Short: "Installs delve",
 		Run: func(cmd *cobra.Command, args []string) {
-			tagFlag := prepareMacnative()
-			execute("go", "install", tagFlag, buildFlags(), DelveMainPackagePath)
+			execute("go", "install", tagFlags(), buildFlags(), DelveMainPackagePath)
 			if runtime.GOOS == "darwin" && os.Getenv("CERT") != "" && canMacnative() {
 				codesign(installedExecutablePath())
 			}
@@ -140,7 +130,7 @@ This option can only be specified if testset is basic or a single package.`)
 
 func checkCert() bool {
 	// If we're on OSX make sure the proper CERT env var is set.
-	if os.Getenv("TRAVIS") == "true" || runtime.GOOS != "darwin" || os.Getenv("CERT") != "" {
+	if runtime.GOOS != "darwin" || os.Getenv("CERT") != "" {
 		return true
 	}
 
@@ -154,7 +144,7 @@ func checkCert() bool {
 		return false
 	}
 	if err != nil {
-		fmt.Printf("An error occoured when generating and installing a new certificate: %v\n", err)
+		fmt.Printf("An error occurred when generating and installing a new certificate: %v\n", err)
 		return false
 	}
 	os.Setenv("CERT", "dlv-cert")
@@ -290,15 +280,37 @@ func prepareMacnative() string {
 	if !checkCert() {
 		return ""
 	}
-	return "-tags=macnative"
+	return "macnative"
+}
+
+func tagFlags() string {
+	var tags []string
+	if mactags := prepareMacnative(); mactags != "" {
+		tags = append(tags, mactags)
+	}
+	if runtime.GOOS == "windows" && runtime.GOARCH == "arm64" {
+		tags = append(tags, "exp.winarm64")
+	}
+	if runtime.GOOS == "linux" && runtime.GOARCH == "ppc64le" {
+		tags = append(tags, "exp.linuxppc64le")
+	}
+	if Tags != nil && len(*Tags) > 0 {
+		tags = append(tags, *Tags...)
+	}
+	if len(tags) == 0 {
+		return ""
+	}
+	return "-tags=" + strings.Join(tags, ",")
 }
 
 func buildFlags() []string {
+	var ldFlags string
 	buildSHA, err := getBuildSHA()
 	if err != nil {
-		log.Fatal(fmt.Errorf("error getting build SHA via git: %w", err))
+		log.Printf("error getting build SHA via git: %v", err)
+	} else {
+		ldFlags = "-X main.Build=" + buildSHA
 	}
-	ldFlags := "-X main.Build=" + buildSHA
 	if runtime.GOOS == "darwin" {
 		ldFlags = "-s " + ldFlags
 	}
@@ -316,9 +328,6 @@ func testFlags() []string {
 	}
 	if NOTimeout {
 		testFlags = append(testFlags, "-timeout", "0")
-	} else if os.Getenv("TRAVIS") == "true" {
-		// Make test timeout shorter than Travis' own timeout so that Go can report which test hangs.
-		testFlags = append(testFlags, "-timeout", "9m")
 	}
 	if len(os.Getenv("TEAMCITY_VERSION")) > 0 {
 		testFlags = append(testFlags, "-json")
@@ -332,20 +341,6 @@ func testFlags() []string {
 func testCmd(cmd *cobra.Command, args []string) {
 	checkCertCmd(nil, nil)
 
-	if os.Getenv("TRAVIS") == "true" && runtime.GOOS == "darwin" {
-		fmt.Println("Building with native backend")
-		execute("go", "build", "-tags=macnative", buildFlags(), DelveMainPackagePath)
-
-		fmt.Println("\nBuilding without native backend")
-		execute("go", "build", buildFlags(), DelveMainPackagePath)
-
-		fmt.Println("\nTesting")
-		os.Setenv("PROCTEST", "lldb")
-		env := []string{}
-		executeq(env, "sudo", "-E", "go", "test", testFlags(), allPackages())
-		return
-	}
-
 	if TestSet == "" && TestBackend == "" && TestBuildMode == "" {
 		if TestRegex != "" {
 			fmt.Printf("Can not use --test-run without --test-set\n")
@@ -386,10 +381,13 @@ func testStandard() {
 		dopie := false
 		switch runtime.GOOS {
 		case "linux":
-			dopie = true
+			if runtime.GOARCH != "ppc64le" {
+				dopie = true
+			}
 		case "windows":
+			// windows/arm64 always uses pie buildmode, no need to test everything again.
 			// only on Go 1.15 or later, with CGO_ENABLED and gcc found in path
-			if goversion.VersionAfterOrEqual(runtime.Version(), 1, 15) {
+			if runtime.GOARCH != "arm64" && goversion.VersionAfterOrEqual(runtime.Version(), 1, 15) {
 				out, err := exec.Command("go", "env", "CGO_ENABLED").CombinedOutput()
 				if err != nil {
 					panic(err)
@@ -400,6 +398,11 @@ func testStandard() {
 					}
 				}
 			}
+		case "darwin":
+			if runtime.GOARCH == "amd64" {
+				// arm64 can only build in pie mode
+				dopie = true
+			}
 		}
 		if dopie {
 			fmt.Println("\nTesting PIE buildmode, default backend")
@@ -443,13 +446,17 @@ func testCmdIntl(testSet, testRegex, testBackend, testBuildMode string) {
 		buildModeFlag = "-test-buildmode=" + testBuildMode
 	}
 
+	env := []string{}
+	if os.Getenv("CI") != "" {
+		env = os.Environ()
+	}
+
 	if len(testPackages) > 3 {
-		env := []string{}
-		executeq(env, "go", "test", testFlags(), buildFlags(), testPackages, backendFlag, buildModeFlag)
+		executeq(env, "go", "test", testFlags(), buildFlags(), tagFlags(), testPackages, backendFlag, buildModeFlag)
 	} else if testRegex != "" {
-		execute("go", "test", testFlags(), buildFlags(), testPackages, "-run="+testRegex, backendFlag, buildModeFlag)
+		executeq(env, "go", "test", testFlags(), buildFlags(), tagFlags(), testPackages, "-run="+testRegex, backendFlag, buildModeFlag)
 	} else {
-		execute("go", "test", testFlags(), buildFlags(), testPackages, backendFlag, buildModeFlag)
+		executeq(env, "go", "test", testFlags(), buildFlags(), tagFlags(), testPackages, backendFlag, buildModeFlag)
 	}
 }
 
@@ -488,7 +495,7 @@ func inpath(exe string) bool {
 
 func allPackages() []string {
 	r := []string{}
-	for _, dir := range strings.Split(getoutput("go", "list", "-mod=vendor", "./..."), "\n") {
+	for _, dir := range strings.Split(getoutput("go", "list", "-mod=vendor", tagFlags(), "./..."), "\n") {
 		dir = strings.TrimSpace(dir)
 		if dir == "" || strings.Contains(dir, "/vendor/") || strings.Contains(dir, "/_scripts") {
 			continue
