@@ -12,6 +12,7 @@ import (
 	"fmt"
 	"go/ast"
 	"go/token"
+	"hash/crc32"
 	"io"
 	"os"
 	"path/filepath"
@@ -21,18 +22,19 @@ import (
 	"sync"
 	"time"
 
+	pdwarf "github.com/go-delve/delve/pkg/dwarf"
 	"github.com/go-delve/delve/pkg/dwarf/frame"
 	"github.com/go-delve/delve/pkg/dwarf/godwarf"
 	"github.com/go-delve/delve/pkg/dwarf/line"
 	"github.com/go-delve/delve/pkg/dwarf/loclist"
 	"github.com/go-delve/delve/pkg/dwarf/op"
 	"github.com/go-delve/delve/pkg/dwarf/reader"
-	"github.com/go-delve/delve/pkg/dwarf/util"
 	"github.com/go-delve/delve/pkg/goversion"
+	"github.com/go-delve/delve/pkg/internal/gosym"
 	"github.com/go-delve/delve/pkg/logflags"
 	"github.com/go-delve/delve/pkg/proc/debuginfod"
+	"github.com/go-delve/gore"
 	"github.com/hashicorp/golang-lru/simplelru"
-	"github.com/sirupsen/logrus"
 )
 
 const (
@@ -50,7 +52,7 @@ type BinaryInfo struct {
 	// GOOS operating system this binary is executing on.
 	GOOS string
 
-	debugInfoDirectories []string
+	DebugInfoDirectories []string
 
 	// BuildID of this binary.
 	BuildID string
@@ -59,8 +61,8 @@ type BinaryInfo struct {
 	Functions []Function
 	// Sources is a list of all source files found in debug_line.
 	Sources []string
-	// LookupFunc maps function names to a description of the function.
-	LookupFunc map[string]*Function
+	// lookupFunc maps function names to a description of the function.
+	lookupFunc map[string][]*Function
 	// lookupGenericFunc maps function names, with their type parameters removed, to functions.
 	// Functions that are not generic are not added to this map.
 	lookupGenericFunc map[string][]*Function
@@ -92,12 +94,8 @@ type BinaryInfo struct {
 	types       map[string]dwarfRef
 	packageVars []packageVar // packageVars is a list of all global/package variables in debug_info, sorted by address
 
-	gStructOffset uint64
-
-	// nameOfRuntimeType maps an address of a runtime._type struct to its
-	// decoded name. Used with versions of Go <= 1.10 to figure out the DIE of
-	// the concrete type of interfaces.
-	nameOfRuntimeType map[uint64]nameOfRuntimeTypeEntry
+	gStructOffset      uint64
+	gStructOffsetIsPtr bool
 
 	// consts[off] lists all the constants with the type defined at offset off.
 	consts constantsMap
@@ -113,7 +111,7 @@ type BinaryInfo struct {
 	// Go 1.17 register ABI is enabled.
 	regabi bool
 
-	logger *logrus.Entry
+	logger logflags.Logger
 }
 
 var (
@@ -131,10 +129,12 @@ var (
 		elf.EM_X86_64:  true,
 		elf.EM_AARCH64: true,
 		elf.EM_386:     true,
+		elf.EM_PPC64:   true,
 	}
 
 	supportedWindowsArch = map[_PEMachine]bool{
 		_IMAGE_FILE_MACHINE_AMD64: true,
+		_IMAGE_FILE_MACHINE_ARM64: true,
 	}
 
 	supportedDarwinArch = map[macho.Cpu]bool{
@@ -280,7 +280,7 @@ func FindFileLocation(p Process, filename string, lineno int) ([]uint64, error)
 	return selectedPCs, nil
 }
 
-// inlRnage is the range of an inlined call
+// inlRange is the range of an inlined call
 type inlRange struct {
 	off   dwarf.Offset
 	depth uint32
@@ -313,8 +313,8 @@ func allInlineCallRanges(tree *godwarf.Tree) []inlRange {
 
 // FindFunction returns the functions with name funcName.
 func (bi *BinaryInfo) FindFunction(funcName string) ([]*Function, error) {
-	if fn := bi.LookupFunc[funcName]; fn != nil {
-		return []*Function{fn}, nil
+	if fns := bi.LookupFunc()[funcName]; fns != nil {
+		return fns, nil
 	}
 	fns := bi.LookupGenericFunc()[funcName]
 	if len(fns) == 0 {
@@ -334,7 +334,7 @@ func FindFunctionLocation(p Process, funcName string, lineOffset int) ([]uint64,
 
 	if lineOffset > 0 {
 		fn := origfns[0]
-		filename, lineno := fn.cu.lineInfo.PCToLine(fn.Entry, fn.Entry)
+		filename, lineno := bi.EntryLineForFunc(fn)
 		return FindFileLocation(p, filename, lineno+lineOffset)
 	}
 
@@ -366,14 +366,16 @@ func FindFunctionLocation(p Process, funcName string, lineOffset int) ([]uint64,
 // If sameline is set FirstPCAfterPrologue will always return an
 // address associated with the same line as fn.Entry.
 func FirstPCAfterPrologue(p Process, fn *Function, sameline bool) (uint64, error) {
-	pc, _, line, ok := fn.cu.lineInfo.PrologueEndPC(fn.Entry, fn.End)
-	if ok {
-		if !sameline {
-			return pc, nil
-		}
-		_, entryLine := fn.cu.lineInfo.PCToLine(fn.Entry, fn.Entry)
-		if entryLine == line {
-			return pc, nil
+	if fn.cu.lineInfo != nil {
+		pc, _, line, ok := fn.cu.lineInfo.PrologueEndPC(fn.Entry, fn.End)
+		if ok {
+			if !sameline {
+				return pc, nil
+			}
+			_, entryLine := p.BinInfo().EntryLineForFunc(fn)
+			if entryLine == line {
+				return pc, nil
+			}
 		}
 	}
 
@@ -382,7 +384,7 @@ func FirstPCAfterPrologue(p Process, fn *Function, sameline bool) (uint64, error
 		return fn.Entry, err
 	}
 
-	if pc == fn.Entry {
+	if pc == fn.Entry && fn.cu.lineInfo != nil {
 		// Look for the first instruction with the stmt flag set, so that setting a
 		// breakpoint with file:line and with the function name always result on
 		// the same instruction being selected.
@@ -394,6 +396,27 @@ func FirstPCAfterPrologue(p Process, fn *Function, sameline bool) (uint64, error
 	return pc, nil
 }
 
+func findRetPC(t *Target, name string) ([]uint64, error) {
+	fn := t.BinInfo().lookupOneFunc(name)
+	if fn == nil {
+		return nil, fmt.Errorf("could not find %s", name)
+	}
+	text, err := Disassemble(t.Memory(), nil, t.Breakpoints(), t.BinInfo(), fn.Entry, fn.End)
+	if err != nil {
+		return nil, err
+	}
+	r := []uint64{}
+	for _, instr := range text {
+		if instr.IsRet() {
+			r = append(r, instr.Loc.PC)
+		}
+	}
+	if len(r) == 0 {
+		return nil, fmt.Errorf("could not find return instruction in %s", name)
+	}
+	return r, nil
+}
+
 // cpuArch is a stringer interface representing CPU architectures.
 type cpuArch interface {
 	String() string
@@ -582,6 +605,25 @@ func (fn *Function) PrologueEndPC() uint64 {
 	return pc
 }
 
+func (fn *Function) AllPCs(excludeFile string, excludeLine int) ([]uint64, error) {
+	if !fn.cu.image.Stripped() {
+		return fn.cu.lineInfo.AllPCsBetween(fn.Entry, fn.End-1, excludeFile, excludeLine)
+	}
+	var pcs []uint64
+	fnFile, lastLine, _ := fn.cu.image.symTable.PCToLine(fn.Entry - fn.cu.image.StaticBase)
+	for pc := fn.Entry - fn.cu.image.StaticBase; pc < fn.End-fn.cu.image.StaticBase; pc++ {
+		f, line, pcfn := fn.cu.image.symTable.PCToLine(pc)
+		if pcfn == nil {
+			continue
+		}
+		if f == fnFile && line > lastLine {
+			lastLine = line
+			pcs = append(pcs, pc+fn.cu.image.StaticBase)
+		}
+	}
+	return pcs, nil
+}
+
 // From $GOROOT/src/runtime/traceback.go:597
 // exportedRuntime reports whether the function is an exported runtime function.
 // It is only for runtime functions, so ASCII A-Z is fine.
@@ -636,7 +678,7 @@ type ElfDynamicSection struct {
 
 // NewBinaryInfo returns an initialized but unloaded BinaryInfo struct.
 func NewBinaryInfo(goos, goarch string) *BinaryInfo {
-	r := &BinaryInfo{GOOS: goos, nameOfRuntimeType: make(map[uint64]nameOfRuntimeTypeEntry), logger: logflags.DebuggerLogger()}
+	r := &BinaryInfo{GOOS: goos, logger: logflags.DebuggerLogger()}
 
 	// TODO: find better way to determine proc arch (perhaps use executable file info).
 	switch goarch {
@@ -646,6 +688,8 @@ func NewBinaryInfo(goos, goarch string) *BinaryInfo {
 		r.Arch = AMD64Arch(goos)
 	case "arm64":
 		r.Arch = ARM64Arch(goos)
+	case "ppc64le":
+		r.Arch = PPC64LEArch(goos)
 	}
 	return r
 }
@@ -657,7 +701,7 @@ func (bi *BinaryInfo) LoadBinaryInfo(path string, entryPoint uint64, debugInfoDi
 		bi.lastModified = fi.ModTime()
 	}
 
-	bi.debugInfoDirectories = debugInfoDirs
+	bi.DebugInfoDirectories = debugInfoDirs
 
 	return bi.AddImage(path, entryPoint)
 }
@@ -679,8 +723,18 @@ func loadBinaryInfo(bi *BinaryInfo, image *Image, path string, entryPoint uint64
 
 // GStructOffset returns the offset of the G
 // struct in thread local storage.
-func (bi *BinaryInfo) GStructOffset() uint64 {
-	return bi.gStructOffset
+func (bi *BinaryInfo) GStructOffset(mem MemoryReadWriter) (uint64, error) {
+	offset := bi.gStructOffset
+	if bi.gStructOffsetIsPtr {
+		// The G struct offset from the TLS section is a pointer
+		// and the address must be dereferenced to find to actual G struct offset.
+		var err error
+		offset, err = readUintRaw(mem, offset, int64(bi.Arch.PtrSize()))
+		if err != nil {
+			return 0, err
+		}
+	}
+	return offset, nil
 }
 
 // LastModified returns the last modified time of the binary.
@@ -690,6 +744,9 @@ func (bi *BinaryInfo) LastModified() time.Time {
 
 // DwarfReader returns a reader for the dwarf data
 func (so *Image) DwarfReader() *reader.Reader {
+	if so.dwarf == nil {
+		return nil
+	}
 	return reader.New(so.dwarf)
 }
 
@@ -702,13 +759,26 @@ func (bi *BinaryInfo) Types() ([]string, error) {
 	return types, nil
 }
 
+func (bi *BinaryInfo) EntryLineForFunc(fn *Function) (string, int) {
+	return bi.pcToLine(fn, fn.Entry)
+}
+
+func (bi *BinaryInfo) pcToLine(fn *Function, pc uint64) (string, int) {
+	if fn.cu.lineInfo == nil {
+		f, l, _ := fn.cu.image.symTable.PCToLine(pc - fn.cu.image.StaticBase)
+		return f, l
+	}
+	f, l := fn.cu.lineInfo.PCToLine(fn.Entry, pc)
+	return f, l
+}
+
 // PCToLine converts an instruction address to a file/line/function.
 func (bi *BinaryInfo) PCToLine(pc uint64) (string, int, *Function) {
 	fn := bi.PCToFunc(pc)
 	if fn == nil {
 		return "", 0, nil
 	}
-	f, ln := fn.cu.lineInfo.PCToLine(fn.Entry, pc)
+	f, ln := bi.pcToLine(fn, pc)
 	return f, ln, fn
 }
 
@@ -781,12 +851,14 @@ type Image struct {
 	debugAddr    *godwarf.DebugAddrSection
 	debugLineStr []byte
 
+	symTable *gosym.Table
+
 	typeCache map[dwarf.Offset]godwarf.Type
 
 	compileUnits []*compileUnit // compileUnits is sorted by increasing DWARF offset
 
-	dwarfTreeCache      *simplelru.LRU
-	runtimeMallocgcTree *godwarf.Tree // patched version of runtime.mallocgc's DIE
+	dwarfTreeCache  *simplelru.LRU
+	workaroundCache map[dwarf.Offset]*godwarf.Tree
 
 	// runtimeTypeToDIE maps between the offset of a runtime._type in
 	// runtime.moduledata.types and the offset of the DIE in debug_info. This
@@ -806,6 +878,10 @@ func (image *Image) registerRuntimeTypeToDIE(entry *dwarf.Entry, ardr *reader.Re
 	}
 }
 
+func (image *Image) Stripped() bool {
+	return image.dwarf == nil
+}
+
 // AddImage adds the specified image to bi, loading data asynchronously.
 // Addr is the relocated entry point for the executable and staticBase (i.e.
 // the relocation offset) for all other images.
@@ -838,7 +914,7 @@ func (bi *BinaryInfo) AddImage(path string, addr uint64) error {
 
 // moduleDataToImage finds the image corresponding to the given module data object.
 func (bi *BinaryInfo) moduleDataToImage(md *moduleData) *Image {
-	fn := bi.PCToFunc(uint64(md.text))
+	fn := bi.PCToFunc(md.text)
 	if fn != nil {
 		return bi.funcToImage(fn)
 	}
@@ -871,6 +947,13 @@ func (bi *BinaryInfo) typeToImage(typ godwarf.Type) *Image {
 	return bi.Images[typ.Common().Index]
 }
 
+func (bi *BinaryInfo) runtimeTypeTypename() string {
+	if goversion.ProducerAfterOrEqual(bi.Producer(), 1, 21) {
+		return "internal/abi.Type"
+	}
+	return "runtime._type"
+}
+
 var errBinaryInfoClose = errors.New("multiple errors closing executable files")
 
 // Close closes all internal readers.
@@ -914,7 +997,7 @@ func (image *Image) Close() error {
 	return err2
 }
 
-func (image *Image) setLoadError(logger *logrus.Entry, fmtstr string, args ...interface{}) {
+func (image *Image) setLoadError(logger logflags.Logger, fmtstr string, args ...interface{}) {
 	image.loadErrMu.Lock()
 	image.loadErr = fmt.Errorf(fmtstr, args...)
 	image.loadErrMu.Unlock()
@@ -929,8 +1012,8 @@ func (image *Image) LoadError() error {
 }
 
 func (image *Image) getDwarfTree(off dwarf.Offset) (*godwarf.Tree, error) {
-	if image.runtimeMallocgcTree != nil && off == image.runtimeMallocgcTree.Offset {
-		return image.runtimeMallocgcTree, nil
+	if image.workaroundCache[off] != nil {
+		return image.workaroundCache[off], nil
 	}
 	if r, ok := image.dwarfTreeCache.Get(off); ok {
 		return r.(*godwarf.Tree), nil
@@ -975,7 +1058,7 @@ func (bi *BinaryInfo) locationExpr(entry godwarf.Entry, attr dwarf.Attr, pc uint
 		return nil, nil, fmt.Errorf("no location attribute %s", attr)
 	}
 	if instr, ok := a.([]byte); ok {
-		return instr, &locationExpr{isBlock: true, instr: instr}, nil
+		return instr, &locationExpr{isBlock: true, instr: instr, regnumToName: bi.Arch.RegnumToString}, nil
 	}
 	off, ok := a.(int64)
 	if !ok {
@@ -985,7 +1068,7 @@ func (bi *BinaryInfo) locationExpr(entry godwarf.Entry, attr dwarf.Attr, pc uint
 	if instr == nil {
 		return nil, nil, fmt.Errorf("could not find loclist entry at %#x for address %#x", off, pc)
 	}
-	return instr, &locationExpr{pc: pc, off: off, instr: instr}, nil
+	return instr, &locationExpr{pc: pc, off: off, instr: instr, regnumToName: bi.Arch.RegnumToString}, nil
 }
 
 type locationExpr struct {
@@ -994,6 +1077,8 @@ type locationExpr struct {
 	off       int64
 	pc        uint64
 	instr     []byte
+
+	regnumToName func(uint64) string
 }
 
 func (le *locationExpr) String() string {
@@ -1004,10 +1089,10 @@ func (le *locationExpr) String() string {
 
 	if le.isBlock {
 		fmt.Fprintf(&descr, "[block] ")
-		op.PrettyPrint(&descr, le.instr)
+		op.PrettyPrint(&descr, le.instr, le.regnumToName)
 	} else {
 		fmt.Fprintf(&descr, "[%#x:%#x] ", le.off, le.pc)
-		op.PrettyPrint(&descr, le.instr)
+		op.PrettyPrint(&descr, le.instr, le.regnumToName)
 	}
 
 	if le.isEscaped {
@@ -1024,7 +1109,7 @@ func (bi *BinaryInfo) LocationCovers(entry *dwarf.Entry, attr dwarf.Attr) ([][2]
 		return nil, fmt.Errorf("attribute %s not found", attr)
 	}
 	if _, isblock := a.([]byte); isblock {
-		return [][2]uint64{[2]uint64{0, ^uint64(0)}}, nil
+		return [][2]uint64{{0, ^uint64(0)}}, nil
 	}
 
 	off, ok := a.(int64)
@@ -1205,34 +1290,105 @@ func (bi *BinaryInfo) parseDebugFrameGeneral(image *Image, debugFrameBytes []byt
 // Alternatively, if the debug file cannot be found be the build-id, Delve
 // will look in directories specified by the debug-info-directories config value.
 func (bi *BinaryInfo) openSeparateDebugInfo(image *Image, exe *elf.File, debugInfoDirectories []string) (*os.File, *elf.File, error) {
-	var debugFilePath string
-	var err error
-	for _, dir := range debugInfoDirectories {
-		var potentialDebugFilePath string
-		if strings.Contains(dir, "build-id") && len(bi.BuildID) > 2 {
-			potentialDebugFilePath = fmt.Sprintf("%s/%s/%s.debug", dir, bi.BuildID[:2], bi.BuildID[2:])
-		} else if strings.HasPrefix(image.Path, "/proc") {
-			path, err := filepath.EvalSymlinks(image.Path)
-			if err == nil {
-				potentialDebugFilePath = fmt.Sprintf("%s/%s.debug", dir, filepath.Base(path))
-			}
-		} else {
-			potentialDebugFilePath = fmt.Sprintf("%s/%s.debug", dir, filepath.Base(image.Path))
+	exePath := image.Path
+	exeName := filepath.Base(image.Path)
+	if strings.HasPrefix(image.Path, "/proc") {
+		var err error
+		exePath, err = filepath.EvalSymlinks(image.Path)
+		if err == nil {
+			exeName = filepath.Base(exePath)
 		}
+	}
+
+	var debugFilePath string
+
+	check := func(potentialDebugFilePath string) bool {
 		_, err := os.Stat(potentialDebugFilePath)
 		if err == nil {
 			debugFilePath = potentialDebugFilePath
-			break
+			return true
+		}
+		return false
+	}
+
+	find := func(f func(string) bool, suffix string) {
+		for _, dir := range debugInfoDirectories {
+			if f != nil && !f(dir) {
+				continue
+			}
+			if check(fmt.Sprintf("%s/%s", dir, suffix)) {
+				break
+			}
 		}
 	}
+
+	if debugFilePath == "" && len(bi.BuildID) > 2 {
+		// Build ID method: look for a file named .build-id/nn/nnnnnnnn.debug in
+		// every debug info directory.
+		find(nil, fmt.Sprintf(".build-id/%s/%s.debug", bi.BuildID[:2], bi.BuildID[2:]))
+	}
+
+	if debugFilePath == "" {
+		// Debug link: method if the executable contains a .gnu_debuglink section
+		// it will look for the file named in the same directory of the
+		// executable, then in a subdirectory named .debug and finally in each
+		// debug info directory in a subdirectory with the same path as the
+		// directory of the executable
+		debugLink, crc := bi.getDebugLink(exe)
+
+		if debugLink != "" {
+			check(filepath.Join(filepath.Dir(exePath), debugLink))
+			if debugFilePath == "" {
+				check(filepath.Join(filepath.Dir(exePath), ".debug", debugLink))
+			}
+			if debugFilePath == "" {
+				suffix := filepath.Join(filepath.Dir(exePath)[1:], debugLink)
+				find(nil, suffix)
+			}
+			if debugFilePath == "" {
+				bi.logger.Warnf("gnu_debuglink link %q not found in any debug info directory", debugLink)
+			}
+		}
+
+		if debugFilePath != "" {
+			// CRC check
+			buf, err := os.ReadFile(debugFilePath)
+			if err == nil {
+				computedCRC := crc32.ChecksumIEEE(buf)
+				if crc != computedCRC {
+					bi.logger.Errorf("gnu_debuglink CRC check failed for %s (want %x got %x)", debugFilePath, crc, computedCRC)
+					debugFilePath = ""
+				}
+
+			}
+		}
+	}
+
+	if debugFilePath == "" && len(bi.BuildID) > 2 {
+		// Previous versions of delve looked for the build id in every debug info
+		// directory that contained the build-id substring. This behavior deviates
+		// from the ones specified by GDB but we keep it for backwards compatibility.
+		find(func(dir string) bool { return strings.Contains(dir, "build-id") }, fmt.Sprintf("%s/%s.debug", bi.BuildID[:2], bi.BuildID[2:]))
+	}
+
+	if debugFilePath == "" {
+		// Previous versions of delve looked for the executable filename (with
+		// .debug extension) in every debug info directory.  This behavior also
+		// deviates from the ones specified by GDB, but we keep it for backwards
+		// compatibility.
+		find(func(dir string) bool { return !strings.Contains(dir, "build-id") }, fmt.Sprintf("%s.debug", exeName))
+	}
+
 	// We cannot find the debug information locally on the system. Try and see if we're on a system that
 	// has debuginfod so that we can use that in order to find any relevant debug information.
 	if debugFilePath == "" {
+		var err error
 		debugFilePath, err = debuginfod.GetDebuginfo(bi.BuildID)
 		if err != nil {
 			return nil, nil, ErrNoDebugInfoFound
 		}
 	}
+
 	sepFile, err := os.OpenFile(debugFilePath, 0, os.ModePerm)
 	if err != nil {
 		return nil, nil, errors.New("can't open separate debug file: " + err.Error())
@@ -1290,13 +1446,69 @@ func loadBinaryInfoElf(bi *BinaryInfo, image *Image, path string, addr uint64, w
 
 	bi.loadBuildID(image, elfFile)
 	var debugInfoBytes []byte
-	image.dwarf, err = elfFile.DWARF()
-	if err != nil {
+	var dwerr error
+	image.dwarf, dwerr = elfFile.DWARF()
+	if dwerr != nil {
 		var sepFile *os.File
 		var serr error
-		sepFile, dwarfFile, serr = bi.openSeparateDebugInfo(image, elfFile, bi.debugInfoDirectories)
+		sepFile, dwarfFile, serr = bi.openSeparateDebugInfo(image, elfFile, bi.DebugInfoDirectories)
 		if serr != nil {
-			return serr
+			if len(bi.Images) <= 1 {
+				fmt.Fprintln(os.Stderr, "Warning: no debug info found, some functionality will be missing such as stack traces and variable evaluation.")
+			}
+			cu := &compileUnit{}
+			cu.image = image
+			symTable, _, err := readPcLnTableElf(elfFile, path)
+			if err != nil {
+				return fmt.Errorf("could not read debug info (%v) and could not read go symbol table (%v)", dwerr, err)
+			}
+			image.symTable = symTable
+			gorefile, err := gore.Open(path)
+			if err != nil {
+				return err
+			}
+			md, err := gorefile.Moduledata()
+			if err != nil {
+				return err
+			}
+			prog := gosym.ProgContaining(elfFile, md.GoFuncValue())
+			inlFuncs := make(map[string]*Function)
+			for _, f := range image.symTable.Funcs {
+				fnEntry := f.Entry + image.StaticBase
+				if prog != nil {
+					inlCalls, err := image.symTable.GetInlineTree(&f, md.GoFuncValue(), prog.Vaddr, prog.ReaderAt)
+					if err != nil {
+						return err
+					}
+					for _, inlfn := range inlCalls {
+						newInlinedCall := InlinedCall{cu: cu, LowPC: fnEntry + uint64(inlfn.ParentPC)}
+						if fn, ok := inlFuncs[inlfn.Name]; ok {
+							fn.InlinedCalls = append(fn.InlinedCalls, newInlinedCall)
+							continue
+						}
+						inlFuncs[inlfn.Name] = &Function{
+							Name:  inlfn.Name,
+							Entry: 0, End: 0,
+							cu: cu,
+							InlinedCalls: []InlinedCall{
+								newInlinedCall,
+							},
+						}
+					}
+				}
+				fn := Function{Name: f.Name, Entry: fnEntry, End: f.End + image.StaticBase, cu: cu}
+				bi.Functions = append(bi.Functions, fn)
+			}
+			for i := range inlFuncs {
+				bi.Functions = append(bi.Functions, *inlFuncs[i])
+			}
+			sort.Sort(functionsDebugInfoByEntry(bi.Functions))
+			for f := range image.symTable.Files {
+				bi.Sources = append(bi.Sources, f)
+			}
+			sort.Strings(bi.Sources)
+			bi.Sources = uniq(bi.Sources)
+			return nil
 		}
 		image.sepDebugCloser = sepFile
 		image.dwarf, err = dwarfFile.DWARF()
@@ -1357,7 +1569,6 @@ func (bi *BinaryInfo) loadSymbolName(image *Image, file *elf.File, wg *sync.Wait
 func (bi *BinaryInfo) loadBuildID(image *Image, file *elf.File) {
 	buildid := file.Section(".note.gnu.build-id")
 	if buildid == nil {
-		bi.logger.Warn("can't find build-id note on binary")
 		return
 	}
 
@@ -1387,6 +1598,28 @@ func (bi *BinaryInfo) loadBuildID(image *Image, file *elf.File) {
 	bi.BuildID = hex.EncodeToString(descBinary)
 }
 
+func (bi *BinaryInfo) getDebugLink(exe *elf.File) (debugLink string, crc uint32) {
+	gnuDebugLink := exe.Section(".gnu_debuglink")
+	if gnuDebugLink == nil {
+		return
+	}
+
+	br := gnuDebugLink.Open()
+	buf, err := io.ReadAll(br)
+	if err != nil {
+		bi.logger.Warnf("can't read .gnu_debuglink: %v", err)
+		return
+	}
+	zero := bytes.Index(buf, []byte{0})
+	if zero <= 0 || len(buf[zero+1:]) < 4 {
+		bi.logger.Warnf("wrong .gnu_debuglink format: %q", buf)
+		return
+	}
+	debugLink = string(buf[:zero])
+	crc = binary.LittleEndian.Uint32(buf[len(buf)-4:])
+	return
+}
+
 func (bi *BinaryInfo) parseDebugFrameElf(image *Image, dwarfFile, exeFile *elf.File, debugInfoBytes []byte, wg *sync.WaitGroup) {
 	defer wg.Done()
 
@@ -1396,7 +1629,12 @@ func (bi *BinaryInfo) parseDebugFrameElf(image *Image, dwarfFile, exeFile *elf.F
 	var ehFrameAddr uint64
 	if ehFrameSection != nil {
 		ehFrameAddr = ehFrameSection.Addr
-		ehFrameData, _ = ehFrameSection.Data()
+		// Workaround for go.dev/cl/429601
+		if ehFrameSection.Type == elf.SHT_NOBITS {
+			ehFrameData = make([]byte, ehFrameSection.Size)
+		} else {
+			ehFrameData, _ = ehFrameSection.Data()
+		}
 	}
 
 	bi.parseDebugFrameGeneral(image, debugFrameData, ".debug_frame", debugFrameErr, ehFrameData, ehFrameAddr, ".eh_frame", frame.DwarfEndian(debugInfoBytes))
@@ -1414,7 +1652,7 @@ func (bi *BinaryInfo) setGStructOffsetElf(image *Image, exe *elf.File, wg *sync.
 	//   emitting runtime.tlsg, a TLS symbol, which is relocated to the chosen
 	//   offset in libc's TLS block.
 	// - On ARM64 (but really, any architecture other than i386 and 86x64) the
-	//   offset is calculate using runtime.tls_g and the formula is different.
+	//   offset is calculated using runtime.tls_g and the formula is different.
 
 	var tls *elf.Prog
 	for _, prog := range exe.Progs {
@@ -1451,13 +1689,16 @@ func (bi *BinaryInfo) setGStructOffsetElf(image *Image, exe *elf.File, wg *sync.
 
 		bi.gStructOffset = tlsg.Value + uint64(bi.Arch.PtrSize()*2) + ((tls.Vaddr - uint64(bi.Arch.PtrSize()*2)) & (tls.Align - 1))
 
+	case elf.EM_PPC64:
+		_ = getSymbol(image, bi.logger, exe, "runtime.tls_g")
+
 	default:
 		// we should never get here
 		panic("architecture not supported")
 	}
 }
 
-func getSymbol(image *Image, logger *logrus.Entry, exe *elf.File, name string) *elf.Symbol {
+func getSymbol(image *Image, logger logflags.Logger, exe *elf.File, name string) *elf.Symbol {
 	symbols, err := exe.Symbols()
 	if err != nil {
 		image.setLoadError(logger, "could not parse ELF symbols: %v", err)
@@ -1496,8 +1737,6 @@ func loadBinaryInfoPE(bi *BinaryInfo, image *Image, path string, entryPoint uint
 	if err != nil {
 		return err
 	}
-
-	//TODO(aarzilli): actually test this when Go supports PIE buildmode on Windows.
 	opth := peFile.OptionalHeader.(*pe.OptionalHeader64)
 	if entryPoint != 0 {
 		image.StaticBase = entryPoint - opth.ImageBase
@@ -1524,16 +1763,55 @@ func loadBinaryInfoPE(bi *BinaryInfo, image *Image, path string, entryPoint uint
 
 	wg.Add(2)
 	go bi.parseDebugFramePE(image, peFile, debugInfoBytes, wg)
-	go bi.loadDebugInfoMaps(image, debugInfoBytes, debugLineBytes, wg, nil)
-
-	// Use ArbitraryUserPointer (0x28) as pointer to pointer
-	// to G struct per:
-	// https://golang.org/src/runtime/cgo/gcc_windows_amd64.c
-
-	bi.gStructOffset = 0x28
+	go bi.loadDebugInfoMaps(image, debugInfoBytes, debugLineBytes, wg, func() {
+		// setGStructOffsetPE requires the image compile units to be loaded,
+		// so it can't be called concurrently with loadDebugInfoMaps.
+		if image.index == 0 {
+			// determine g struct offset only when loading the executable file.
+			bi.setGStructOffsetPE(entryPoint, peFile)
+		}
+	})
 	return nil
 }
 
+func (bi *BinaryInfo) setGStructOffsetPE(entryPoint uint64, peFile *pe.File) {
+	readtls_g := func() uint64 {
+		for _, s := range peFile.Symbols {
+			if s.Name == "runtime.tls_g" {
+				i := int(s.SectionNumber) - 1
+				if 0 <= i && i < len(peFile.Sections) {
+					sect := peFile.Sections[i]
+					if s.Value < sect.VirtualSize {
+						return entryPoint + uint64(sect.VirtualAddress) + uint64(s.Value)
+					}
+				}
+				break
+			}
+		}
+		return 0
+	}
+	switch _PEMachine(peFile.Machine) {
+	case _IMAGE_FILE_MACHINE_AMD64:
+		producer := bi.Producer()
+		if producer != "" && goversion.ProducerAfterOrEqual(producer, 1, 20) {
+			// Use runtime.tls_g as pointer to offset from GS to G struct:
+			// https://golang.org/src/runtime/sys_windows_amd64.s
+			bi.gStructOffset = readtls_g()
+			bi.gStructOffsetIsPtr = true
+		} else {
+			// Use ArbitraryUserPointer (0x28) as pointer to pointer
+			// to G struct per:
+			// https://golang.org/src/runtime/cgo/gcc_windows_amd64.c
+			bi.gStructOffset = 0x28
+		}
+	case _IMAGE_FILE_MACHINE_ARM64:
+		// Use runtime.tls_g as pointer to offset from R18 to G struct:
+		// https://golang.org/src/runtime/sys_windows_arm64.s
+		bi.gStructOffset = readtls_g()
+		bi.gStructOffsetIsPtr = true
+	}
+}
+
 func openExecutablePathPE(path string) (*pe.File, io.Closer, error) {
 	f, err := os.OpenFile(path, 0, os.ModePerm)
 	if err != nil {
@@ -1565,20 +1843,82 @@ func loadBinaryInfoMacho(bi *BinaryInfo, image *Image, path string, entryPoint u
 	}
 
 	if entryPoint != 0 {
-		// This is a little bit hacky. We use the entryPoint variable, but it
-		// actually holds the address of the mach-o header. We can use this
-		// to calculate the offset to the non-aslr location of the mach-o header
-		// (which is 0x100000000)
-		image.StaticBase = entryPoint - 0x100000000
+		machoOff := uint64(0x100000000)
+		for _, ld := range exe.Loads {
+			if seg, _ := ld.(*macho.Segment); seg != nil {
+				if seg.Name == "__TEXT" {
+					machoOff = seg.Addr
+					break
+				}
+			}
+		}
+		logflags.DebuggerLogger().Debugf("entryPoint %#x machoOff %#x", entryPoint, machoOff)
+		image.StaticBase = entryPoint - machoOff
 	}
 
 	image.closer = exe
 	if !supportedDarwinArch[exe.Cpu] {
 		return &ErrUnsupportedArch{os: "darwin", cpuArch: exe.Cpu}
 	}
-	image.dwarf, err = exe.DWARF()
-	if err != nil {
-		return err
+	var dwerr error
+	image.dwarf, dwerr = exe.DWARF()
+	if dwerr != nil {
+		if len(bi.Images) <= 1 {
+			fmt.Fprintln(os.Stderr, "Warning: no debug info found, some functionality will be missing such as stack traces and variable evaluation.")
+		}
+		symTable, err := readPcLnTableMacho(exe, path)
+		if err != nil {
+			return fmt.Errorf("could not read debug info (%v) and could not read go symbol table (%v)", dwerr, err)
+		}
+		image.symTable = symTable
+		cu := &compileUnit{}
+		cu.image = image
+		gorefile, err := gore.Open(path)
+		if err != nil {
+			return err
+		}
+		md, err := gorefile.Moduledata()
+		if err != nil {
+			return err
+		}
+		seg := gosym.SegmentContaining(exe, md.GoFuncValue())
+		inlFuncs := make(map[string]*Function)
+		for _, f := range image.symTable.Funcs {
+			fnEntry := f.Entry + image.StaticBase
+			if seg != nil {
+				inlCalls, err := image.symTable.GetInlineTree(&f, md.GoFuncValue(), seg.Addr, seg.ReaderAt)
+				if err != nil {
+					return err
+				}
+				for _, inlfn := range inlCalls {
+					newInlinedCall := InlinedCall{cu: cu, LowPC: fnEntry + uint64(inlfn.ParentPC)}
+					if fn, ok := inlFuncs[inlfn.Name]; ok {
+						fn.InlinedCalls = append(fn.InlinedCalls, newInlinedCall)
+						continue
+					}
+					inlFuncs[inlfn.Name] = &Function{
+						Name:  inlfn.Name,
+						Entry: 0, End: 0,
+						cu: cu,
+						InlinedCalls: []InlinedCall{
+							newInlinedCall,
+						},
+					}
+				}
+			}
+			fn := Function{Name: f.Name, Entry: fnEntry, End: f.End + image.StaticBase, cu: cu}
+			bi.Functions = append(bi.Functions, fn)
+		}
+		for i := range inlFuncs {
+			bi.Functions = append(bi.Functions, *inlFuncs[i])
+		}
+		sort.Sort(functionsDebugInfoByEntry(bi.Functions))
+		for f := range image.symTable.Files {
+			bi.Sources = append(bi.Sources, f)
+		}
+		sort.Strings(bi.Sources)
+		bi.Sources = uniq(bi.Sources)
+		return nil
 	}
 	debugInfoBytes, err := godwarf.GetDebugSectionMacho(exe, "info")
 	if err != nil {
@@ -1643,8 +1983,7 @@ func (bi *BinaryInfo) parseDebugFrameMacho(image *Image, exe *macho.File, debugI
 //
 // [golang/go#25841]: https://github.com/golang/go/issues/25841
 func (bi *BinaryInfo) macOSDebugFrameBugWorkaround() {
-	//TODO: log extensively because of bugs in the field
-	if bi.GOOS != "darwin" || bi.Arch.Name != "arm64" {
+	if bi.GOOS != "darwin" {
 		return
 	}
 	if len(bi.Images) > 1 {
@@ -1657,9 +1996,28 @@ func (bi *BinaryInfo) macOSDebugFrameBugWorkaround() {
 	if !ok {
 		return
 	}
-	if exe.Flags&macho.FlagPIE == 0 {
-		bi.logger.Infof("debug_frame workaround not needed: not a PIE (%#x)", exe.Flags)
-		return
+	if bi.Arch.Name == "arm64" {
+		if exe.Flags&macho.FlagPIE == 0 {
+			bi.logger.Infof("debug_frame workaround not needed: not a PIE (%#x)", exe.Flags)
+			return
+		}
+	} else {
+		prod := goversion.ParseProducer(bi.Producer())
+		if !prod.AfterOrEqual(goversion.GoVersion{Major: 1, Minor: 19, Rev: 3}) && !prod.IsDevel() {
+			bi.logger.Infof("debug_frame workaround not needed (version %q on %s)", bi.Producer(), bi.Arch.Name)
+			return
+		}
+		found := false
+		for i := range bi.frameEntries {
+			if bi.frameEntries[i].CIE.CIE_id == ^uint32(0) && bi.frameEntries[i].Begin() < 0x4000000 {
+				found = true
+				break
+			}
+		}
+		if !found {
+			bi.logger.Infof("debug_frame workaround not needed (all FDEs above 0x4000000)")
+			return
+		}
 	}
 
 	// Find first Go function (first = lowest entry point)
@@ -1715,6 +2073,8 @@ func (bi *BinaryInfo) macOSDebugFrameBugWorkaround() {
 
 // Do not call this function directly it isn't able to deal correctly with package paths
 func (bi *BinaryInfo) findType(name string) (godwarf.Type, error) {
+	name = strings.ReplaceAll(name, "interface{", "interface {")
+	name = strings.ReplaceAll(name, "struct{", "struct {")
 	ref, found := bi.types[name]
 	if !found {
 		return nil, reader.ErrTypeNotFound
@@ -1745,7 +2105,7 @@ func (bi *BinaryInfo) findTypeExpr(expr ast.Expr) (godwarf.Type, error) {
 	}
 	bi.expandPackagesInType(expr)
 	if snode, ok := expr.(*ast.StarExpr); ok {
-		// Pointer types only appear in the dwarf informations when
+		// Pointer types only appear in the dwarf information when
 		// a pointer to the type is used in the target program, here
 		// we create a pointer type on the fly so that the user can
 		// specify a pointer to any variable used in the target program
@@ -1843,7 +2203,7 @@ func (bi *BinaryInfo) loadDebugInfoMaps(image *Image, debugInfoBytes, debugLineB
 
 	image.runtimeTypeToDIE = make(map[uint64]runtimeTypeDIE)
 
-	ctxt := newLoadDebugInfoMapsContext(bi, image, util.ReadUnitVersions(debugInfoBytes))
+	ctxt := newLoadDebugInfoMapsContext(bi, image, pdwarf.ReadUnitVersions(debugInfoBytes))
 
 	reader := image.DwarfReader()
 
@@ -1883,11 +2243,7 @@ func (bi *BinaryInfo) loadDebugInfoMaps(image *Image, debugInfoBytes, debugLineB
 			if hasLineInfo && lineInfoOffset >= 0 && lineInfoOffset < int64(len(debugLineBytes)) {
 				var logfn func(string, ...interface{})
 				if logflags.DebugLineErrors() {
-					logger := logrus.New().WithFields(logrus.Fields{"layer": "dwarf-line"})
-					logger.Logger.Level = logrus.DebugLevel
-					logfn = func(fmt string, args ...interface{}) {
-						logger.Printf(fmt, args)
-					}
+					logfn = logflags.DebugLineLogger().Debugf
 				}
 				cu.lineInfo = line.Parse(compdir, bytes.NewBuffer(debugLineBytes[lineInfoOffset:]), image.debugLineStr, logfn, image.StaticBase, bi.GOOS == "windows", bi.Arch.PtrSize())
 			}
@@ -1910,7 +2266,7 @@ func (bi *BinaryInfo) loadDebugInfoMaps(image *Image, debugInfoBytes, debugLineB
 			}
 			gopkg, _ := entry.Val(godwarf.AttrGoPackageName).(string)
 			if cu.isgo && gopkg != "" {
-				bi.PackageMap[gopkg] = append(bi.PackageMap[gopkg], escapePackagePath(strings.Replace(cu.name, "\\", "/", -1)))
+				bi.PackageMap[gopkg] = append(bi.PackageMap[gopkg], escapePackagePath(strings.ReplaceAll(cu.name, "\\", "/")))
 			}
 			image.compileUnits = append(image.compileUnits, cu)
 			if entry.Children {
@@ -1930,11 +2286,8 @@ func (bi *BinaryInfo) loadDebugInfoMaps(image *Image, debugInfoBytes, debugLineB
 	sort.Sort(functionsDebugInfoByEntry(bi.Functions))
 	sort.Sort(packageVarsByAddr(bi.packageVars))
 
-	bi.LookupFunc = make(map[string]*Function)
+	bi.lookupFunc = nil
 	bi.lookupGenericFunc = nil
-	for i := range bi.Functions {
-		bi.LookupFunc[bi.Functions[i].Name] = &bi.Functions[i]
-	}
 
 	for _, cu := range image.compileUnits {
 		if cu.lineInfo != nil {
@@ -1946,28 +2299,12 @@ func (bi *BinaryInfo) loadDebugInfoMaps(image *Image, debugInfoBytes, debugLineB
 	sort.Strings(bi.Sources)
 	bi.Sources = uniq(bi.Sources)
 
-	if bi.regabi {
-		// prepare patch for runtime.mallocgc's DIE
-		fn := bi.LookupFunc["runtime.mallocgc"]
-		if fn != nil && fn.cu.image == image {
-			tree, err := image.getDwarfTree(fn.offset)
-			if err == nil {
-				tree.Children, err = regabiMallocgcWorkaround(bi)
-				if err != nil {
-					bi.logger.Errorf("could not patch runtime.mallogc: %v", err)
-				} else {
-					image.runtimeMallocgcTree = tree
-				}
-			}
-		}
-	}
-
 	if cont != nil {
 		cont()
 	}
 }
 
-// LookupGenericFunc returns a map that allows searching for instantiations of generic function by specificying a function name without type parameters.
+// LookupGenericFunc returns a map that allows searching for instantiations of generic function by specifying a function name without type parameters.
 // For example the key "pkg.(*Receiver).Amethod" will find all instantiations of Amethod:
 //   - pkg.(*Receiver[.shape.int]).Amethod
 //   - pkg.(*Receiver[.shape.*uint8]).Amethod
@@ -1985,6 +2322,25 @@ func (bi *BinaryInfo) LookupGenericFunc() map[string][]*Function {
 	return bi.lookupGenericFunc
 }
 
+func (bi *BinaryInfo) LookupFunc() map[string][]*Function {
+	if bi.lookupFunc == nil {
+		bi.lookupFunc = make(map[string][]*Function)
+		for i := range bi.Functions {
+			name := bi.Functions[i].Name
+			bi.lookupFunc[name] = append(bi.lookupFunc[name], &bi.Functions[i])
+		}
+	}
+	return bi.lookupFunc
+}
+
+func (bi *BinaryInfo) lookupOneFunc(name string) *Function {
+	fns := bi.LookupFunc()[name]
+	if fns == nil {
+		return nil
+	}
+	return fns[0]
+}
+
 // loadDebugInfoMapsCompileUnit loads entry from a single compile unit.
 func (bi *BinaryInfo) loadDebugInfoMapsCompileUnit(ctxt *loadDebugInfoMapsContext, image *Image, reader *reader.Reader, cu *compileUnit) {
 	hasAttrGoPkgName := goversion.ProducerAfterOrEqual(cu.producer, 1, 13)
@@ -2031,7 +2387,7 @@ func (bi *BinaryInfo) loadDebugInfoMapsCompileUnit(ctxt *loadDebugInfoMapsContex
 				var addr uint64
 				if loc, ok := entry.Val(dwarf.AttrLocation).([]byte); ok {
 					if len(loc) == bi.Arch.PtrSize()+1 && op.Opcode(loc[0]) == op.DW_OP_addr {
-						addr, _ = util.ReadUintRaw(bytes.NewReader(loc[1:]), binary.LittleEndian, bi.Arch.PtrSize())
+						addr, _ = pdwarf.ReadUintRaw(bytes.NewReader(loc[1:]), binary.LittleEndian, bi.Arch.PtrSize())
 					}
 				}
 				if !cu.isgo {
@@ -2266,7 +2622,7 @@ func (bi *BinaryInfo) loadDebugInfoMapsInlinedCalls(ctxt *loadDebugInfoMapsConte
 }
 
 func uniq(s []string) []string {
-	if len(s) <= 0 {
+	if len(s) == 0 {
 		return s
 	}
 	src, dst := 1, 1
@@ -2329,7 +2685,7 @@ func escapePackagePath(pkg string) string {
 	if slash < 0 {
 		slash = 0
 	}
-	return pkg[:slash] + strings.Replace(pkg[slash:], ".", "%2e", -1)
+	return pkg[:slash] + strings.ReplaceAll(pkg[slash:], ".", "%2e")
 }
 
 // Looks up symbol (either functions or global variables) at address addr.
@@ -2380,7 +2736,7 @@ func (bi *BinaryInfo) ListPackagesBuildInfo(includeFiles bool) []*PackageBuildIn
 			continue
 		}
 
-		ip := strings.Replace(cu.name, "\\", "/", -1)
+		ip := strings.ReplaceAll(cu.name, "\\", "/")
 		if _, ok := m[ip]; !ok {
 			path := cu.lineInfo.FirstFile()
 			if ext := filepath.Ext(path); ext != ".go" && ext != ".s" {
