@@ -1,8 +1,10 @@
 package native
 
 import (
+	"errors"
 	"os"
 	"runtime"
+	"time"
 
 	"github.com/go-delve/delve/pkg/proc"
 )
@@ -24,11 +26,11 @@ type nativeProcess struct {
 	// Thread used to read and write memory
 	memthread *nativeThread
 
-	os             *osProcessDetails
-	firstStart     bool
-	ptraceChan     chan func()
-	ptraceDoneChan chan interface{}
-	childProcess   bool // this process was launched, not attached to
+	os           *osProcessDetails
+	firstStart   bool
+	ptraceThread *ptraceThread
+	childProcess bool // this process was launched, not attached to
+	followExec   bool // automatically attach to new processes
 
 	// Controlling terminal file descriptor for
 	// this process.
@@ -45,19 +47,47 @@ type nativeProcess struct {
 // `handlePtraceFuncs`.
 func newProcess(pid int) *nativeProcess {
 	dbp := &nativeProcess{
-		pid:            pid,
-		threads:        make(map[int]*nativeThread),
-		breakpoints:    proc.NewBreakpointMap(),
-		firstStart:     true,
-		os:             new(osProcessDetails),
-		ptraceChan:     make(chan func()),
-		ptraceDoneChan: make(chan interface{}),
-		bi:             proc.NewBinaryInfo(runtime.GOOS, runtime.GOARCH),
+		pid:          pid,
+		threads:      make(map[int]*nativeThread),
+		breakpoints:  proc.NewBreakpointMap(),
+		firstStart:   true,
+		os:           new(osProcessDetails),
+		ptraceThread: newPtraceThread(),
+		bi:           proc.NewBinaryInfo(runtime.GOOS, runtime.GOARCH),
 	}
-	go dbp.handlePtraceFuncs()
 	return dbp
 }
 
+// newChildProcess is like newProcess but uses the same ptrace thread as dbp.
+func newChildProcess(dbp *nativeProcess, pid int) *nativeProcess {
+	return &nativeProcess{
+		pid:          pid,
+		threads:      make(map[int]*nativeThread),
+		breakpoints:  proc.NewBreakpointMap(),
+		firstStart:   true,
+		os:           new(osProcessDetails),
+		ptraceThread: dbp.ptraceThread.acquire(),
+		bi:           proc.NewBinaryInfo(runtime.GOOS, runtime.GOARCH),
+	}
+}
+
+// WaitFor waits for a process as specified by waitFor.
+func WaitFor(waitFor *proc.WaitFor) (int, error) {
+	t0 := time.Now()
+	seen := make(map[int]struct{})
+	for (waitFor.Duration == 0) || (time.Since(t0) < waitFor.Duration) {
+		pid, err := waitForSearchProcess(waitFor.Name, seen)
+		if err != nil {
+			return 0, err
+		}
+		if pid != 0 {
+			return pid, nil
+		}
+		time.Sleep(waitFor.Interval)
+	}
+	return 0, errors.New("waitfor duration expired")
+}
+
 // BinInfo will return the binary info struct associated with this process.
 func (dbp *nativeProcess) BinInfo() *proc.BinaryInfo {
 	return dbp.bi
@@ -66,17 +96,24 @@ func (dbp *nativeProcess) BinInfo() *proc.BinaryInfo {
 // StartCallInjection notifies the backend that we are about to inject a function call.
 func (dbp *nativeProcess) StartCallInjection() (func(), error) { return func() {}, nil }
 
+// detachWithoutGroup is a helper function to detach from a process which we
+// haven't added to a process group yet.
+func detachWithoutGroup(dbp *nativeProcess, kill bool) error {
+	grp := &processGroup{procs: []*nativeProcess{dbp}}
+	return grp.Detach(dbp.pid, kill)
+}
+
 // Detach from the process being debugged, optionally killing it.
-func (dbp *nativeProcess) Detach(kill bool) (err error) {
-	if dbp.exited {
+func (procgrp *processGroup) Detach(pid int, kill bool) (err error) {
+	dbp := procgrp.procForPid(pid)
+	if ok, _ := dbp.Valid(); !ok {
 		return nil
 	}
 	if kill && dbp.childProcess {
-		err := dbp.kill()
+		err := procgrp.kill(dbp)
 		if err != nil {
 			return err
 		}
-		dbp.bi.Close()
 		return nil
 	}
 	dbp.execPtraceFunc(func() {
@@ -133,8 +170,8 @@ func (dbp *nativeProcess) Breakpoints() *proc.BreakpointMap {
 // RequestManualStop sets the `manualStopRequested` flag and
 // sends SIGSTOP to all threads.
 func (dbp *nativeProcess) RequestManualStop(cctx *proc.ContinueOnceContext) error {
-	if dbp.exited {
-		return proc.ErrProcessExited{Pid: dbp.pid}
+	if ok, err := dbp.Valid(); !ok {
+		return err
 	}
 	return dbp.requestManualStop()
 }
@@ -172,21 +209,77 @@ func (dbp *nativeProcess) EraseBreakpoint(bp *proc.Breakpoint) error {
 	return dbp.memthread.clearSoftwareBreakpoint(bp)
 }
 
-// ContinueOnce will continue the target until it stops.
-// This could be the result of a breakpoint or signal.
-func (dbp *nativeProcess) ContinueOnce(cctx *proc.ContinueOnceContext) (proc.Thread, proc.StopReason, error) {
-	if dbp.exited {
-		return nil, proc.StopExited, proc.ErrProcessExited{Pid: dbp.pid}
+type processGroup struct {
+	procs     []*nativeProcess
+	addTarget proc.AddTargetFunc
+}
+
+func (procgrp *processGroup) numValid() int {
+	n := 0
+	for _, p := range procgrp.procs {
+		if ok, _ := p.Valid(); ok {
+			n++
+		}
+	}
+	return n
+}
+
+func (procgrp *processGroup) procForThread(tid int) *nativeProcess {
+	for _, p := range procgrp.procs {
+		if p.threads[tid] != nil {
+			return p
+		}
+	}
+	return nil
+}
+
+func (procgrp *processGroup) procForPid(pid int) *nativeProcess {
+	for _, p := range procgrp.procs {
+		if p.pid == pid {
+			return p
+		}
+	}
+	return nil
+}
+
+func (procgrp *processGroup) add(p *nativeProcess, pid int, currentThread proc.Thread, path string, stopReason proc.StopReason, cmdline string) (*proc.Target, error) {
+	tgt, err := procgrp.addTarget(p, pid, currentThread, path, stopReason, cmdline)
+	if tgt == nil {
+		i := len(procgrp.procs)
+		procgrp.procs = append(procgrp.procs, p)
+		procgrp.detachChild(p)
+		if i == len(procgrp.procs)-1 {
+			procgrp.procs = procgrp.procs[:i]
+		}
+	}
+	if err != nil {
+		return nil, err
+	}
+	if tgt != nil {
+		procgrp.procs = append(procgrp.procs, p)
+	}
+	return tgt, nil
+}
+
+func (procgrp *processGroup) ContinueOnce(cctx *proc.ContinueOnceContext) (proc.Thread, proc.StopReason, error) {
+	if len(procgrp.procs) != 1 && runtime.GOOS != "linux" && runtime.GOOS != "windows" {
+		panic("not implemented")
+	}
+	if procgrp.numValid() == 0 {
+		return nil, proc.StopExited, proc.ErrProcessExited{Pid: procgrp.procs[0].pid}
 	}
 
 	for {
-
-		if err := dbp.resume(); err != nil {
+		err := procgrp.resume()
+		if err != nil {
 			return nil, proc.StopUnknown, err
 		}
-
-		for _, th := range dbp.threads {
-			th.CurrentBreakpoint.Clear()
+		for _, dbp := range procgrp.procs {
+			if valid, _ := dbp.Valid(); valid {
+				for _, th := range dbp.threads {
+					th.CurrentBreakpoint.Clear()
+				}
+			}
 		}
 
 		if cctx.ResumeChan != nil {
@@ -194,16 +287,29 @@ func (dbp *nativeProcess) ContinueOnce(cctx *proc.ContinueOnceContext) (proc.Thr
 			cctx.ResumeChan = nil
 		}
 
-		trapthread, err := dbp.trapWait(-1)
+		trapthread, err := trapWait(procgrp, -1)
 		if err != nil {
 			return nil, proc.StopUnknown, err
 		}
-		trapthread, err = dbp.stop(cctx, trapthread)
+		trapthread, err = procgrp.stop(cctx, trapthread)
 		if err != nil {
 			return nil, proc.StopUnknown, err
 		}
 		if trapthread != nil {
+			dbp := procgrp.procForThread(trapthread.ID)
 			dbp.memthread = trapthread
+			// refresh memthread for every other process
+			for _, p2 := range procgrp.procs {
+				if p2.exited || p2.detached || p2 == dbp {
+					continue
+				}
+				for _, th := range p2.threads {
+					p2.memthread = th
+					if th.SoftExc() {
+						break
+					}
+				}
+			}
 			return trapthread, proc.StopUnknown, nil
 		}
 	}
@@ -224,48 +330,61 @@ func (dbp *nativeProcess) FindBreakpoint(pc uint64, adjustPC bool) (*proc.Breakp
 	return nil, false
 }
 
-// initialize will ensure that all relevant information is loaded
-// so the process is ready to be debugged.
-func (dbp *nativeProcess) initialize(path string, debugInfoDirs []string) (*proc.Target, error) {
-	if err := initialize(dbp); err != nil {
-		return nil, err
+func (dbp *nativeProcess) initializeBasic() (string, error) {
+	cmdline, err := initialize(dbp)
+	if err != nil {
+		return "", err
 	}
 	if err := dbp.updateThreadList(); err != nil {
+		return "", err
+	}
+	return cmdline, nil
+}
+
+// initialize will ensure that all relevant information is loaded
+// so the process is ready to be debugged.
+func (dbp *nativeProcess) initialize(path string, debugInfoDirs []string) (*proc.TargetGroup, error) {
+	cmdline, err := dbp.initializeBasic()
+	if err != nil {
 		return nil, err
 	}
 	stopReason := proc.StopLaunched
 	if !dbp.childProcess {
 		stopReason = proc.StopAttached
 	}
-	tgt, err := proc.NewTarget(dbp, dbp.pid, dbp.memthread, proc.NewTargetConfig{
-		Path:          path,
+	procgrp := &processGroup{}
+	grp, addTarget := proc.NewGroup(procgrp, proc.NewTargetGroupConfig{
 		DebugInfoDirs: debugInfoDirs,
 
 		// We disable asyncpreempt for the following reasons:
 		//  - on Windows asyncpreempt is incompatible with debuggers, see:
 		//    https://github.com/golang/go/issues/36494
-		//  - freebsd's backend is generally broken and asyncpreempt makes it even more so, see:
-		//    https://github.com/go-delve/delve/issues/1754
 		//  - on linux/arm64 asyncpreempt can sometimes restart a sequence of
 		//    instructions, if the sequence happens to contain a breakpoint it will
 		//    look like the breakpoint was hit twice when it was "logically" only
 		//    executed once.
 		//    See: https://go-review.googlesource.com/c/go/+/208126
-		DisableAsyncPreempt: runtime.GOOS == "windows" || runtime.GOOS == "freebsd" || (runtime.GOOS == "linux" && runtime.GOARCH == "arm64"),
+		//	- on linux/ppc64le according to @laboger, they had issues in the past
+		//	  with gdb once AsyncPreempt was enabled. While implementing the port,
+		//	  few tests failed while it was enabled, but cannot be warrantied that
+		//	  disabling it fixed the issues.
+		DisableAsyncPreempt: runtime.GOOS == "windows" || (runtime.GOOS == "linux" && runtime.GOARCH == "arm64") || (runtime.GOOS == "linux" && runtime.GOARCH == "ppc64le"),
 
 		StopReason: stopReason,
-		CanDump:    runtime.GOOS == "linux" || runtime.GOOS == "windows",
+		CanDump:    runtime.GOOS == "linux" || runtime.GOOS == "freebsd" || (runtime.GOOS == "windows" && runtime.GOARCH == "amd64"),
 	})
+	procgrp.addTarget = addTarget
+	tgt, err := procgrp.add(dbp, dbp.pid, dbp.memthread, path, stopReason, cmdline)
 	if err != nil {
 		return nil, err
 	}
-	if dbp.bi.Arch.Name == "arm64" {
+	if dbp.bi.Arch.Name == "arm64" || dbp.bi.Arch.Name == "ppc64le" {
 		dbp.iscgo = tgt.IsCgo()
 	}
-	return tgt, nil
+	return grp, nil
 }
 
-func (dbp *nativeProcess) handlePtraceFuncs() {
+func (pt *ptraceThread) handlePtraceFuncs() {
 	// We must ensure here that we are running on the same thread during
 	// while invoking the ptrace(2) syscall. This is due to the fact that ptrace(2) expects
 	// all commands after PTRACE_ATTACH to come from the same thread.
@@ -278,21 +397,21 @@ func (dbp *nativeProcess) handlePtraceFuncs() {
 		defer runtime.UnlockOSThread()
 	}
 
-	for fn := range dbp.ptraceChan {
+	for fn := range pt.ptraceChan {
 		fn()
-		dbp.ptraceDoneChan <- nil
+		pt.ptraceDoneChan <- nil
 	}
+	close(pt.ptraceDoneChan)
 }
 
 func (dbp *nativeProcess) execPtraceFunc(fn func()) {
-	dbp.ptraceChan <- fn
-	<-dbp.ptraceDoneChan
+	dbp.ptraceThread.ptraceChan <- fn
+	<-dbp.ptraceThread.ptraceDoneChan
 }
 
 func (dbp *nativeProcess) postExit() {
 	dbp.exited = true
-	close(dbp.ptraceChan)
-	close(dbp.ptraceDoneChan)
+	dbp.ptraceThread.release()
 	dbp.bi.Close()
 	if dbp.ctty != nil {
 		dbp.ctty.Close()
@@ -305,11 +424,11 @@ func (dbp *nativeProcess) writeSoftwareBreakpoint(thread *nativeThread, addr uin
 	return err
 }
 
-func openRedirects(redirects [3]string, foreground bool) (stdin, stdout, stderr *os.File, closefn func(), err error) {
+func openRedirects(stdinPath string, stdoutOR proc.OutputRedirect, stderrOR proc.OutputRedirect, foreground bool) (stdin, stdout, stderr *os.File, closefn func(), err error) {
 	toclose := []*os.File{}
 
-	if redirects[0] != "" {
-		stdin, err = os.Open(redirects[0])
+	if stdinPath != "" {
+		stdin, err = os.Open(stdinPath)
 		if err != nil {
 			return nil, nil, nil, nil, err
 		}
@@ -318,24 +437,29 @@ func openRedirects(redirects [3]string, foreground bool) (stdin, stdout, stderr
 		stdin = os.Stdin
 	}
 
-	create := func(path string, dflt *os.File) *os.File {
-		if path == "" {
-			return dflt
+	create := func(redirect proc.OutputRedirect, dflt *os.File) (f *os.File) {
+		if redirect.Path != "" {
+			f, err = os.Create(redirect.Path)
+			if f != nil {
+				toclose = append(toclose, f)
+			}
+
+			return f
+		} else if redirect.File != nil {
+			toclose = append(toclose, redirect.File)
+
+			return redirect.File
 		}
-		var f *os.File
-		f, err = os.Create(path)
-		if f != nil {
-			toclose = append(toclose, f)
-		}
-		return f
+
+		return dflt
 	}
 
-	stdout = create(redirects[1], os.Stdout)
+	stdout = create(stdoutOR, os.Stdout)
 	if err != nil {
 		return nil, nil, nil, nil, err
 	}
 
-	stderr = create(redirects[2], os.Stderr)
+	stderr = create(stderrOR, os.Stderr)
 	if err != nil {
 		return nil, nil, nil, nil, err
 	}
@@ -348,3 +472,31 @@ func openRedirects(redirects [3]string, foreground bool) (stdin, stdout, stderr
 
 	return stdin, stdout, stderr, closefn, nil
 }
+
+type ptraceThread struct {
+	ptraceRefCnt   int
+	ptraceChan     chan func()
+	ptraceDoneChan chan interface{}
+}
+
+func newPtraceThread() *ptraceThread {
+	pt := &ptraceThread{
+		ptraceChan:     make(chan func()),
+		ptraceDoneChan: make(chan interface{}),
+		ptraceRefCnt:   1,
+	}
+	go pt.handlePtraceFuncs()
+	return pt
+}
+
+func (pt *ptraceThread) acquire() *ptraceThread {
+	pt.ptraceRefCnt++
+	return pt
+}
+
+func (pt *ptraceThread) release() {
+	pt.ptraceRefCnt--
+	if pt.ptraceRefCnt == 0 {
+		close(pt.ptraceChan)
+	}
+}
