@@ -9,6 +9,7 @@ import (
 	"go/constant"
 	"go/token"
 	"math"
+	"math/bits"
 	"reflect"
 	"sort"
 	"strconv"
@@ -50,7 +51,7 @@ const (
 	FloatIsNormal floatSpecial = iota
 	// FloatIsNaN means the float is a special NaN value.
 	FloatIsNaN
-	// FloatIsPosInf means the float is a special positive inifitiy value.
+	// FloatIsPosInf means the float is a special positive infinity value.
 	FloatIsPosInf
 	// FloatIsNegInf means the float is a special negative infinity value.
 	FloatIsNegInf
@@ -81,7 +82,7 @@ const (
 	// the variable is the return value of a function call and allocated on a
 	// frame that no longer exists)
 	VariableFakeAddress
-	// VariableCPrt means the variable is a C pointer
+	// VariableCPtr means the variable is a C pointer
 	VariableCPtr
 	// VariableCPURegister means this variable is a CPU register.
 	VariableCPURegister
@@ -124,6 +125,9 @@ type Variable struct {
 	// number of elements to skip when loading a map
 	mapSkip int
 
+	// Children lists the variables sub-variables. What constitutes a child
+	// depends on the variable's type. For pointers, there's one child
+	// representing the pointed-to variable.
 	Children []Variable
 
 	loaded     bool
@@ -196,7 +200,7 @@ const (
 // G represents a runtime G (goroutine) structure (at least the
 // fields that Delve is interested in).
 type G struct {
-	ID      int    // Goroutine ID
+	ID      int64  // Goroutine ID
 	PC      uint64 // PC of goroutine when it was parked.
 	SP      uint64 // SP of goroutine when it was parked.
 	BP      uint64 // BP of goroutine when it was parked (go >= 1.7).
@@ -312,7 +316,7 @@ func GoroutinesInfo(dbp *Target, start, count int) ([]*G, int, error) {
 	}
 
 	var (
-		threadg = map[int]*G{}
+		threadg = map[int64]*G{}
 		allg    []*G
 	)
 
@@ -367,7 +371,7 @@ func GoroutinesInfo(dbp *Target, start, count int) ([]*G, int, error) {
 
 // FindGoroutine returns a G struct representing the goroutine
 // specified by `gid`.
-func FindGoroutine(dbp *Target, gid int) (*G, error) {
+func FindGoroutine(dbp *Target, gid int64) (*G, error) {
 	if selg := dbp.SelectedGoroutine(); (gid == -1) || (selg != nil && selg.ID == gid) || (selg == nil && gid == 0) {
 		// Return the currently selected goroutine in the following circumstances:
 		//
@@ -438,8 +442,12 @@ func getGVariable(thread Thread) (*Variable, error) {
 
 	gaddr, hasgaddr := regs.GAddr()
 	if !hasgaddr {
-		var err error
-		gaddr, err = readUintRaw(thread.ProcessMemory(), regs.TLS()+thread.BinInfo().GStructOffset(), int64(thread.BinInfo().Arch.PtrSize()))
+		bi := thread.BinInfo()
+		offset, err := bi.GStructOffset(thread.ProcessMemory())
+		if err != nil {
+			return nil, err
+		}
+		gaddr, err = readUintRaw(thread.ProcessMemory(), regs.TLS()+offset, int64(bi.Arch.PtrSize()))
 		if err != nil {
 			return nil, err
 		}
@@ -490,7 +498,7 @@ func (g *G) Defer() *Defer {
 // UserCurrent returns the location the users code is at,
 // or was at before entering a runtime function.
 func (g *G) UserCurrent() Location {
-	it, err := g.stackIterator(0)
+	it, err := goroutineStackIterator(nil, g, 0)
 	if err != nil {
 		return g.CurrentLoc
 	}
@@ -528,7 +536,7 @@ func (g *G) StartLoc(tgt *Target) Location {
 	if fn == nil {
 		return Location{PC: g.StartPC}
 	}
-	f, l := fn.cu.lineInfo.PCToLine(fn.Entry, fn.Entry)
+	f, l := tgt.BinInfo().EntryLineForFunc(fn)
 	return Location{PC: fn.Entry, File: f, Line: l, Fn: fn}
 }
 
@@ -662,9 +670,9 @@ func newVariable(name string, addr uint64, dwarfType godwarf.Type, bi *BinaryInf
 	case *godwarf.StringType:
 		v.Kind = reflect.String
 		v.stride = 1
-		v.fieldType = &godwarf.UintType{BasicType: godwarf.BasicType{CommonType: godwarf.CommonType{ByteSize: 1, Name: "byte"}, BitSize: 8, BitOffset: 0}}
+		v.fieldType = &godwarf.UintType{BasicType: godwarf.BasicType{CommonType: godwarf.CommonType{ByteSize: 1, Name: "byte", ReflectKind: reflect.Uint8}, BitSize: 8, BitOffset: 0}}
 		if v.Addr != 0 {
-			v.Base, v.Len, v.Unreadable = readStringInfo(v.mem, v.bi.Arch, v.Addr)
+			v.Base, v.Len, v.Unreadable = readStringInfo(v.mem, v.bi.Arch, v.Addr, t)
 		}
 	case *godwarf.SliceType:
 		v.Kind = reflect.Slice
@@ -833,7 +841,7 @@ var ErrUnreadableG = errors.New("could not read G struct")
 
 func (v *Variable) parseG() (*G, error) {
 	mem := v.mem
-	gaddr := uint64(v.Addr)
+	gaddr := v.Addr
 	_, deref := v.RealType.(*godwarf.PtrType)
 
 	if deref {
@@ -870,8 +878,8 @@ func (v *Variable) parseG() (*G, error) {
 	if bpvar := schedVar.fieldVariable("bp"); /* +rtype -opt uintptr */ bpvar != nil && bpvar.Value != nil {
 		bp, _ = constant.Int64Val(bpvar.Value)
 	}
-	if bpvar := schedVar.fieldVariable("lr"); /* +rtype -opt uintptr */ bpvar != nil && bpvar.Value != nil {
-		lr, _ = constant.Int64Val(bpvar.Value)
+	if lrvar := schedVar.fieldVariable("lr"); /* +rtype -opt uintptr */ lrvar != nil && lrvar.Value != nil {
+		lr, _ = constant.Int64Val(lrvar.Value)
 	}
 
 	unreadable := false
@@ -886,7 +894,17 @@ func (v *Variable) parseG() (*G, error) {
 		return n
 	}
 
-	id := loadInt64Maybe("goid")             // +rtype int64
+	loadUint64Maybe := func(name string) uint64 {
+		vv := v.loadFieldNamed(name)
+		if vv == nil {
+			unreadable = true
+			return 0
+		}
+		n, _ := constant.Uint64Val(vv.Value)
+		return n
+	}
+
+	id := loadUint64Maybe("goid")            // +rtype int64|uint64
 	gopc := loadInt64Maybe("gopc")           // +rtype uintptr
 	startpc := loadInt64Maybe("startpc")     // +rtype uintptr
 	waitSince := loadInt64Maybe("waitsince") // +rtype int64
@@ -896,15 +914,34 @@ func (v *Variable) parseG() (*G, error) {
 	}
 	var stackhi, stacklo uint64
 	if stackVar := v.loadFieldNamed("stack"); /* +rtype stack */ stackVar != nil {
-		if stackhiVar := stackVar.fieldVariable("hi"); /* +rtype uintptr */ stackhiVar != nil {
+		if stackhiVar := stackVar.fieldVariable("hi"); /* +rtype uintptr */ stackhiVar != nil && stackhiVar.Value != nil {
 			stackhi, _ = constant.Uint64Val(stackhiVar.Value)
+		} else {
+			unreadable = true
 		}
-		if stackloVar := stackVar.fieldVariable("lo"); /* +rtype uintptr */ stackloVar != nil {
+		if stackloVar := stackVar.fieldVariable("lo"); /* +rtype uintptr */ stackloVar != nil && stackloVar.Value != nil {
 			stacklo, _ = constant.Uint64Val(stackloVar.Value)
+		} else {
+			unreadable = true
 		}
 	}
 
-	status := loadInt64Maybe("atomicstatus") // +rtype uint32
+	status := uint64(0)
+	if atomicStatus := v.loadFieldNamed("atomicstatus"); /* +rtype uint32|runtime/internal/atomic.Uint32 */ atomicStatus != nil {
+		if constant.Val(atomicStatus.Value) != nil {
+			status, _ = constant.Uint64Val(atomicStatus.Value)
+		} else {
+			atomicStatus := atomicStatus              // +rtype runtime/internal/atomic.Uint32
+			vv := atomicStatus.fieldVariable("value") // +rtype uint32
+			if vv == nil {
+				unreadable = true
+			} else {
+				status, _ = constant.Uint64Val(vv.Value)
+			}
+		}
+	} else {
+		unreadable = true
+	}
 
 	if unreadable {
 		return nil, ErrUnreadableG
@@ -915,14 +952,14 @@ func (v *Variable) parseG() (*G, error) {
 	v.Name = "runtime.curg"
 
 	g := &G{
-		ID:         int(id),
+		ID:         int64(id),
 		GoPC:       uint64(gopc),
 		StartPC:    uint64(startpc),
 		PC:         uint64(pc),
 		SP:         uint64(sp),
 		BP:         uint64(bp),
 		LR:         uint64(lr),
-		Status:     uint64(status),
+		Status:     status,
 		WaitSince:  waitSince,
 		WaitReason: waitReason,
 		CurrentLoc: Location{PC: uint64(pc), File: f, Line: l, Fn: fn},
@@ -1152,19 +1189,19 @@ func extractVarInfoFromEntry(tgt *Target, bi *BinaryInfo, image *Image, regs op.
 		return nil, err
 	}
 
-	t, err = resolveParametricType(tgt, bi, mem, t, dictAddr)
+	t, err = resolveParametricType(bi, mem, t, dictAddr)
 	if err != nil {
 		// Log the error, keep going with t, which will be the shape type
-		logflags.DebuggerLogger().Errorf("could not resolve parametric type of %s", n)
+		logflags.DebuggerLogger().Errorf("could not resolve parametric type of %s: %v", n, err)
 	}
 
 	addr, pieces, descr, err := bi.Location(entry, dwarf.AttrLocation, regs.PC(), regs, mem)
 	if pieces != nil {
 		var cmem *compositeMemory
 		if tgt != nil {
-			addr, cmem, err = tgt.newCompositeMemory(mem, regs, pieces, descr)
+			addr, cmem, err = tgt.newCompositeMemory(mem, regs, pieces, descr, t.Common().ByteSize)
 		} else {
-			cmem, err = newCompositeMemory(mem, bi.Arch, regs, pieces)
+			cmem, err = newCompositeMemory(mem, bi.Arch, regs, pieces, t.Common().ByteSize)
 			if cmem != nil {
 				cmem.base = fakeAddressUnresolv
 				addr = int64(cmem.base)
@@ -1211,6 +1248,40 @@ func (v *Variable) maybeDereference() *Variable {
 	}
 }
 
+// loadPtr assumes that v is a pointer and loads its value. v also gets a child
+// variable, representing the pointed-to value. If v is already loaded,
+// loadPtr() is a no-op.
+func (v *Variable) loadPtr() {
+	if len(v.Children) > 0 {
+		// We've already loaded this variable.
+		return
+	}
+
+	t := v.RealType.(*godwarf.PtrType)
+	v.Len = 1
+
+	var child *Variable
+	if v.Unreadable == nil {
+		ptrval, err := readUintRaw(v.mem, v.Addr, t.ByteSize)
+		if err == nil {
+			child = v.newVariable("", ptrval, t.Type, DereferenceMemory(v.mem))
+		} else {
+			// We failed to read the pointer value; mark v as unreadable.
+			v.Unreadable = err
+		}
+	}
+
+	if v.Unreadable != nil {
+		// Pointers get a child even if their value can't be read, to
+		// maintain backwards compatibility.
+		child = v.newVariable("", 0 /* addr */, t.Type, DereferenceMemory(v.mem))
+		child.Unreadable = fmt.Errorf("parent pointer unreadable: %w", v.Unreadable)
+	}
+
+	v.Children = []Variable{*child}
+	v.Value = constant.MakeUint64(v.Children[0].Addr)
+}
+
 func loadValues(vars []*Variable, cfg LoadConfig) {
 	for i := range vars {
 		vars[i].loadValueInternal(0, cfg)
@@ -1230,16 +1301,26 @@ func (v *Variable) loadValueInternal(recurseLevel int, cfg LoadConfig) {
 	v.loaded = true
 	switch v.Kind {
 	case reflect.Ptr, reflect.UnsafePointer:
-		v.Len = 1
-		v.Children = []Variable{*v.maybeDereference()}
+		v.loadPtr()
 		if cfg.FollowPointers {
 			// Don't increase the recursion level when dereferencing pointers
 			// unless this is a pointer to interface (which could cause an infinite loop)
 			nextLvl := recurseLevel
+			checkLvl := false
 			if v.Children[0].Kind == reflect.Interface {
 				nextLvl++
+			} else if ptyp, isptr := v.RealType.(*godwarf.PtrType); isptr {
+				_, elemTypIsPtr := resolveTypedef(ptyp.Type).(*godwarf.PtrType)
+				if elemTypIsPtr {
+					nextLvl++
+					checkLvl = true
+				}
+			}
+			if checkLvl && recurseLevel > cfg.MaxVariableRecurse {
+				v.Children[0].OnlyAddr = true
+			} else {
+				v.Children[0].loadValueInternal(nextLvl, cfg)
 			}
-			v.Children[0].loadValueInternal(nextLvl, cfg)
 		} else {
 			v.Children[0].OnlyAddr = true
 		}
@@ -1353,13 +1434,13 @@ func (v *Variable) loadValueInternal(recurseLevel int, cfg LoadConfig) {
 	case reflect.Func:
 		v.readFunctionPtr()
 	default:
-		v.Unreadable = fmt.Errorf("unknown or unsupported kind: \"%s\"", v.Kind.String())
+		v.Unreadable = fmt.Errorf("unknown or unsupported kind: %q", v.Kind.String())
 	}
 }
 
 // convertToEface converts srcv into an "interface {}" and writes it to
 // dstv.
-// Dstv must be a variable of type "inteface {}" and srcv must either be an
+// Dstv must be a variable of type "interface {}" and srcv must either be an
 // interface or a pointer shaped variable (map, channel, pointer or struct
 // containing a single pointer)
 func convertToEface(srcv, dstv *Variable) error {
@@ -1373,7 +1454,7 @@ func convertToEface(srcv, dstv *Variable) error {
 			return srcv.Unreadable
 		}
 		_type = _type.maybeDereference()
-		dstv.writeEmptyInterface(uint64(_type.Addr), data)
+		dstv.writeEmptyInterface(_type.Addr, data)
 		return nil
 	}
 	typeAddr, typeKind, runtimeTypeFound, err := dwarfToRuntimeType(srcv.bi, srcv.mem, srcv.RealType)
@@ -1386,30 +1467,38 @@ func convertToEface(srcv, dstv *Variable) error {
 	return dstv.writeEmptyInterface(typeAddr, srcv)
 }
 
-func readStringInfo(mem MemoryReadWriter, arch *Arch, addr uint64) (uint64, int64, error) {
+func readStringInfo(mem MemoryReadWriter, arch *Arch, addr uint64, typ *godwarf.StringType) (uint64, int64, error) {
 	// string data structure is always two ptrs in size. Addr, followed by len
 	// http://research.swtch.com/godata
 
 	mem = cacheMemory(mem, addr, arch.PtrSize()*2)
 
-	// read len
-	strlen, err := readIntRaw(mem, addr+uint64(arch.PtrSize()), int64(arch.PtrSize()))
-	if err != nil {
-		return 0, 0, fmt.Errorf("could not read string len %s", err)
-	}
-	if strlen < 0 {
-		return 0, 0, fmt.Errorf("invalid length: %d", strlen)
+	var strlen int64
+	var outaddr uint64
+	var err error
+
+	for _, field := range typ.StructType.Field {
+		switch field.Name {
+		case "len":
+			strlen, err = readIntRaw(mem, addr+uint64(field.ByteOffset), int64(arch.PtrSize()))
+			if err != nil {
+				return 0, 0, fmt.Errorf("could not read string len %s", err)
+			}
+			if strlen < 0 {
+				return 0, 0, fmt.Errorf("invalid length: %d", strlen)
+			}
+		case "str":
+			outaddr, err = readUintRaw(mem, addr+uint64(field.ByteOffset), int64(arch.PtrSize()))
+			if err != nil {
+				return 0, 0, fmt.Errorf("could not read string pointer %s", err)
+			}
+			if addr == 0 {
+				return 0, 0, nil
+			}
+		}
 	}
 
-	// read addr
-	addr, err = readUintRaw(mem, addr, int64(arch.PtrSize()))
-	if err != nil {
-		return 0, 0, fmt.Errorf("could not read string pointer %s", err)
-	}
-	if addr == 0 {
-		return 0, 0, nil
-	}
-	return addr, strlen, nil
+	return outaddr, strlen, nil
 }
 
 func readStringValue(mem MemoryReadWriter, addr uint64, strlen int64, cfg LoadConfig) (string, error) {
@@ -1587,12 +1676,19 @@ func (v *Variable) loadArrayValues(recurseLevel int, cfg LoadConfig) {
 		v.Unreadable = errors.New("Negative array length")
 		return
 	}
+	if v.Base == 0 && v.Len > 0 {
+		v.Unreadable = errors.New("non-zero length array with nil base")
+		return
+	}
 
 	count := v.Len
 	// Cap number of elements
 	if count > int64(cfg.MaxArrayValues) {
 		count = int64(cfg.MaxArrayValues)
 	}
+	if v.Base+uint64(v.stride*count) < v.Base {
+		v.Unreadable = fmt.Errorf("bad array base address %#x", v.Base)
+	}
 
 	if v.stride < maxArrayStridePrefetch {
 		v.mem = cacheMemory(v.mem, v.Base, int(v.stride*count))
@@ -1632,7 +1728,7 @@ func (v *Variable) readComplex(size int64) {
 		return
 	}
 
-	ftyp := &godwarf.FloatType{BasicType: godwarf.BasicType{CommonType: godwarf.CommonType{ByteSize: fs, Name: fmt.Sprintf("float%d", fs)}, BitSize: fs * 8, BitOffset: 0}}
+	ftyp := godwarf.FakeBasicType("float", int(fs*8))
 
 	realvar := v.newVariable("real", v.Addr, ftyp, v.mem)
 	imagvar := v.newVariable("imaginary", v.Addr+uint64(fs), ftyp, v.mem)
@@ -1642,13 +1738,13 @@ func (v *Variable) readComplex(size int64) {
 }
 
 func (v *Variable) writeComplex(real, imag float64, size int64) error {
-	err := v.writeFloatRaw(real, int64(size/2))
+	err := v.writeFloatRaw(real, size/2)
 	if err != nil {
 		return err
 	}
 	imagaddr := *v
 	imagaddr.Addr += uint64(size / 2)
-	return imagaddr.writeFloatRaw(imag, int64(size/2))
+	return imagaddr.writeFloatRaw(imag, size/2)
 }
 
 func readIntRaw(mem MemoryReadWriter, addr uint64, size int64) (int64, error) {
@@ -1685,7 +1781,7 @@ func (v *Variable) writeUint(value uint64, size int64) error {
 	case 4:
 		binary.LittleEndian.PutUint32(val, uint32(value))
 	case 8:
-		binary.LittleEndian.PutUint64(val, uint64(value))
+		binary.LittleEndian.PutUint64(val, value)
 	}
 
 	_, err := v.mem.WriteMemory(v.Addr, val)
@@ -1709,7 +1805,7 @@ func readUintRaw(mem MemoryReadWriter, addr uint64, size int64) (uint64, error)
 	case 4:
 		n = uint64(binary.LittleEndian.Uint32(val))
 	case 8:
-		n = uint64(binary.LittleEndian.Uint64(val))
+		n = binary.LittleEndian.Uint64(val)
 	}
 
 	return n, nil
@@ -1745,7 +1841,7 @@ func (v *Variable) writeFloatRaw(f float64, size int64) error {
 		n := float32(f)
 		binary.Write(buf, binary.LittleEndian, n)
 	case 8:
-		n := float64(f)
+		n := f
 		binary.Write(buf, binary.LittleEndian, n)
 	}
 
@@ -1766,7 +1862,7 @@ func (v *Variable) writeZero() error {
 	return err
 }
 
-// writeInterface writes the empty interface of type typeAddr and data as the data field.
+// writeEmptyInterface writes the empty interface of type typeAddr and data as the data field.
 func (v *Variable) writeEmptyInterface(typeAddr uint64, data *Variable) error {
 	dstType, dstData, _ := v.readInterface()
 	if v.Unreadable != nil {
@@ -1782,7 +1878,7 @@ func (v *Variable) writeSlice(len, cap int64, base uint64) error {
 		switch f.Name {
 		case sliceArrayFieldName:
 			arrv, _ := v.toField(f)
-			if err := arrv.writeUint(uint64(base), arrv.RealType.Size()); err != nil {
+			if err := arrv.writeUint(base, arrv.RealType.Size()); err != nil {
 				return err
 			}
 		case sliceLenFieldName:
@@ -1801,8 +1897,8 @@ func (v *Variable) writeSlice(len, cap int64, base uint64) error {
 }
 
 func (v *Variable) writeString(len, base uint64) error {
-	writePointer(v.bi, v.mem, uint64(v.Addr), base)
-	writePointer(v.bi, v.mem, uint64(v.Addr)+uint64(v.bi.Arch.PtrSize()), len)
+	writePointer(v.bi, v.mem, v.Addr, base)
+	writePointer(v.bi, v.mem, v.Addr+uint64(v.bi.Arch.PtrSize()), len)
 	return nil
 }
 
@@ -1835,7 +1931,7 @@ func (v *Variable) readFunctionPtr() {
 	}
 
 	v.Base = val
-	fn := v.bi.PCToFunc(uint64(v.Base))
+	fn := v.bi.PCToFunc(v.Base)
 	if fn == nil {
 		v.Unreadable = fmt.Errorf("could not find function for %#v", v.Base)
 		return
@@ -2175,7 +2271,7 @@ func (v *Variable) readInterface() (_type, data *Variable, isnil bool) {
 
 	// +rtype -field iface.tab *itab
 	// +rtype -field iface.data unsafe.Pointer
-	// +rtype -field eface._type *_type
+	// +rtype -field eface._type *_type|*internal/abi.Type
 	// +rtype -field eface.data unsafe.Pointer
 
 	for _, f := range ityp.Field {
@@ -2186,7 +2282,7 @@ func (v *Variable) readInterface() (_type, data *Variable, isnil bool) {
 			isnil = tab.Addr == 0
 			if !isnil {
 				var err error
-				_type, err = tab.structMember("_type") // +rtype *_type
+				_type, err = tab.structMember("_type") // +rtype *_type|*internal/abi.Type
 				if err != nil {
 					v.Unreadable = fmt.Errorf("invalid interface type: %v", err)
 					return
@@ -2260,7 +2356,7 @@ func (v *Variable) ConstDescr() string {
 	}
 	if typename := v.DwarfType.Common().Name; !strings.Contains(typename, ".") || strings.HasPrefix(typename, "C.") {
 		// only attempt to use constants for user defined type, otherwise every
-		// int variable with value 1 will be described with os.SEEK_CUR and other
+		// int variable with value 1 will be described with io.SeekCurrent and other
 		// similar problems.
 		return ""
 	}
@@ -2310,7 +2406,7 @@ func (v *Variable) registerVariableTypeConv(newtyp string) (*Variable, error) {
 			child.Kind = reflect.Uint32
 			n = 4
 		case "uint64":
-			child = newConstant(constant.MakeUint64(uint64(binary.LittleEndian.Uint64(v.reg.Bytes[i:]))), v.mem)
+			child = newConstant(constant.MakeUint64(binary.LittleEndian.Uint64(v.reg.Bytes[i:])), v.mem)
 			child.Kind = reflect.Uint64
 			n = 8
 		case "float32":
@@ -2333,7 +2429,7 @@ func (v *Variable) registerVariableTypeConv(newtyp string) (*Variable, error) {
 						break
 					}
 				}
-				if n == 0 || popcnt(uint64(n)) != 1 {
+				if n == 0 || bits.OnesCount64(uint64(n)) != 1 {
 					return nil, fmt.Errorf("unknown CPU register type conversion to %q", newtyp)
 				}
 				n = n / 8
@@ -2352,23 +2448,6 @@ func (v *Variable) registerVariableTypeConv(newtyp string) (*Variable, error) {
 	return v, nil
 }
 
-// popcnt is the number of bits set to 1 in x.
-// It's the same as math/bits.OnesCount64, copied here so that we can build
-// on versions of go that don't have math/bits.
-func popcnt(x uint64) int {
-	const m0 = 0x5555555555555555 // 01010101 ...
-	const m1 = 0x3333333333333333 // 00110011 ...
-	const m2 = 0x0f0f0f0f0f0f0f0f // 00001111 ...
-	const m = 1<<64 - 1
-	x = x>>1&(m0&m) + x&(m0&m)
-	x = x>>2&(m1&m) + x&(m1&m)
-	x = (x>>4 + x) & (m2 & m)
-	x += x >> 8
-	x += x >> 16
-	x += x >> 32
-	return int(x) & (1<<7 - 1)
-}
-
 func isCgoType(bi *BinaryInfo, typ godwarf.Type) bool {
 	cu := bi.Images[typ.Common().Index].findCompileUnitForOffset(typ.Common().Offset)
 	if cu == nil {
@@ -2411,7 +2490,7 @@ func (cm constantsMap) Get(typ godwarf.Type) *constantType {
 		sort.Sort(constantValuesByValue(ctyp.values))
 		for i := range ctyp.values {
 			ctyp.values[i].name = strings.TrimPrefix(ctyp.values[i].name, typepkg)
-			if popcnt(uint64(ctyp.values[i].value)) == 1 {
+			if bits.OnesCount64(uint64(ctyp.values[i].value)) == 1 {
 				ctyp.values[i].singleBit = true
 			}
 		}
@@ -2522,6 +2601,10 @@ func (v *Variable) formatTime() {
 	} else {
 		// the full signed 64-bit wall seconds since Jan 1 year 1 is stored in ext
 		var t time.Time
+		if ext > int64(maxAddSeconds/time.Second)*1000 {
+			// avoid doing the add loop below if it will take too much time
+			return
+		}
 		for ext > int64(maxAddSeconds/time.Second) {
 			t = t.Add(maxAddSeconds)
 			ext -= int64(maxAddSeconds / time.Second)
